#include <signal.h>
#include <string.h>
#include "sim.h"
#include "table.h"

void message(str, par1, par2, par3, par4, par5)
      /* Write a message in the status line */
     char *str;
     int par1, par2, par3, par4, par5;
{
 static char buf[128];
 sprintf(buf, str, par1, par2, par3, par4, par5);
 if (!using_x)
 {
  if (is_tty)
   printf("%s\n", buf);
 }
#ifndef SIM_MAIN
 else
 {
  int len= strlen(buf)+1;
  write(comms_sock[1], "m", 1);
  write(comms_sock[1], &len, sizeof(int));
  write(comms_sock[1], buf, strlen(buf)+1);
 }
#endif
}

char radix= 'd';
int unsignd= 0;

char *NumOut(int N, char radix, int unsignd)
{
 static char str[40];
 switch(radix)
 {
  case 'd': sprintf(str, unsignd ? "%-10u" : "%-10d", N); break;
  case 'x': sprintf(str, "%-8x", N); break;
  case 'o': sprintf(str, "%-11o", N); break;
  default: sprintf(str, "HELP! Unknown number format '%c'\n", radix);
 }
 return str;
} 

char *NumOutNoPad(int N, int unsignd, int mask, char radix)
{
 static char str[40];
 if (radix == 'x' || radix == 'o' || unsignd)
  N &= mask;  /* The number may be only a certain number of bits long */
 switch(radix)
 {
  case 'd': sprintf(str, unsignd ? "%u" : "%d", N); break;
  case 'x': sprintf(str, "%x", N); break;
  case 'o': sprintf(str, "%o", N); break;
  default: sprintf(str, "HELP! Unknown number format '%c'\n", radix);
 }
 return str;
} 

char *dis(int instr, char radix)
{
 /*
  * *** WARNING *** After calling this, rs1, rs2 etc may be corrupted.
  * You MUST call parse on the instruction you want to execute after
  * calling this routine. 
  */
 static char buf[80];
 int type, op;
 int loadOp= 0, storeOp= 0;
 int unsignd= 0;
 op= (instr >> 26) & 0x3f;
 type= opTable[op].format;
 if (type == IFMT)
 {
  loadOp= (op >= 0x20 && op < 0x28);  /* Load & store instrs have different
					    format */
  storeOp= (op >= 0x28 && op < 0x30);
 }
 if (opTable[op].opCode == SPECIAL)
  op= specialTable[instr & 0x3f];
 else
  op= opTable[op].opCode;
 if (type == IFMT)
  unsignd= (  op == OP_ADDUI || op == OP_SUBUI
           || op == OP_SEQUI || op == OP_SNEUI
	   || op == OP_SLTUI || op == OP_SGTUI
           || op == OP_SLEUI || op == OP_SGEUI);
 if (op > maxInstrOp) /* Unknown instruction */
  sprintf(buf, "???");
 else
 {
  parse(type, instr);
  if (type == RFMT)
    sprintf(buf, "%s r%d, r%d, r%d", operationNames[op], rd, rs1, rs2);
  else if (type == IFMT)
   if (loadOp)
    sprintf(buf, "%s %r%d, %s(r%d)", operationNames[op],
	    rd, NumOutNoPad(extra, 0, 0xffff, radix), rs1);
   else if (storeOp)
    sprintf(buf, "%s %s(r%d), r%d", operationNames[op],
	    NumOutNoPad(extra, 0, 0xffff, radix), rs1, rd);
   else if (op == OP_BEQZ || op == OP_BNEZ)
    sprintf(buf, "%s r%d, %s", operationNames[op], rs1,
	    NumOutNoPad(extra, 0, 0xffff, radix));
   else if (op == OP_TRAP)
    sprintf(buf, "%s %s", operationNames[op], NumOutNoPad(extra, 1, 0xffff, radix));
   else if (op == OP_JR || op == OP_JALR)
    sprintf(buf, "%s r%d", operationNames[op], rs1);
   else if (op == OP_SLLI || op == OP_SRLI || op == OP_SRAI)
    sprintf(buf, "%s r%d, r%d, %s", operationNames[op], rd, rs1,
	    NumOutNoPad(extra, 1, 0x1f, radix));
   else if (unsignd)
     sprintf(buf, "%s r%d, r%d, %s", operationNames[op], rd, rs1,
	     NumOutNoPad(extra, 1, 0xffff, radix));
   else
    sprintf(buf, "%s r%d, r%d, %s", operationNames[op], rd, rs1,
	     NumOutNoPad(extra, 0, 0xffff, radix));
  else /* type == JFMT */
   sprintf(buf, "%s %s", operationNames[op], NumOutNoPad(extra, 0, 0x3ffffff, radix));
 }
 return buf;
}

int getNum(char *str, int *result, char radix)
{
 static int success;
 switch (radix)
 {
  case 'd': success= sscanf(str, "%d", result); break;
  case 'x': success= sscanf(str, "%x", result); break;
  case 'o': success= sscanf(str, "%o", result); break;
  default: printf("HELP!  Invalid radix!\n"); return -1;
 }
 return success;
}

int Read(int fd, char *dst, int num)
{
 int result= read(fd, dst, num);
 if (result <= 0) exit(0);  /* Parent has exited */
 if (result != num)
 {
  message("PANIC: Read: expected %d bytes, got %d", num, result);
  exit(1);
 }
 return num;
}

void stop()
     /* Return control to X interface */
{
 int mask;
 static char str[80];
 char *ptr;
 int atpc= 0;
 void (*handler)();

 handler= signal(ATTN_SIG, SIG_IGN);  /* Remember what it was for later */
 gotInt= 0;
 ignoreBpt= 1;  /* So we can read & write mem without bpt problems */
 if (!using_x)
 {
  if (is_tty)
  {
   if (FetchMem(&atpc, pc) == 1 /* Bad error */)
    printf("pc=%x  *pc=<error>  r1=%d r2=%d r3=%d r4=%d r5=%d r6=%d\n", pc, regs[1], regs[2], regs[3], regs[4], regs[5], regs[6]);
   else
    printf("%-20s   pc=%x  *pc=%x  r1=%d r2=%d r3=%d r4=%d r5=%d r6=%d\n", dis(atpc, radix), pc, atpc, regs[1], regs[2], regs[3], regs[4], regs[5], regs[6]);
  }
 }
#ifndef SIM_MAIN
 else
 {
  write(comms_sock[1], "s", 1);
 }
#endif

 for (;;)
 {
  if (!using_x)
  {
   if (is_tty) printf("sim> ");
   if (gets(str) == NULL) { if (is_tty) puts("quit"); exit(0); }
   if (str[0] == 0) continue;  /* Blank line */
   ptr= strtok(str, " \t");
   if (!ptr) continue;  /* Just spaces */
  /* Commands:
     c[ont]
     s[tep]
     r[eg] [num [value]]
     r[eg] [pc [value]]
     b[reak] [location]
     b[reak] delete number
     rb[reak] [location]
     rb[reak] delete number
     wb[reak] [location]
     wb[reak] delete number
     di[s] [start [end]]
     ra[dix] radix
     help
     quit
   */
     
   if (!strcmp(ptr, "c") || !strcmp(ptr, "cont"))
    break;
   else if (!strcmp(ptr, "s") || !strcmp(ptr, "step"))
   {
    gotInt= 1;  /* Simulate single step by pretending we got an interrupt */
    break;
   }
   else if (!strcmp(ptr, "r") || !strcmp(ptr, "reg"))
   {
    int i;
    ptr= strtok(NULL, " \t");
    if (!ptr)  /* Print out all regs */
    {
     for (i=0; i < 31; i += 4)
     {
      /* Must do these as separate print statements, as NumOut has a static
	 string return value. */
      printf("r%02d= %s", i, NumOut(regs[i], radix, unsignd));
      printf("r%02d= %s", i+1, NumOut(regs[i+1], radix, unsignd));
      printf("r%02d= %s", i+2, NumOut(regs[i+2], radix, unsignd));
      printf("r%02d= %s\n", i+3, NumOut(regs[i+3], radix, unsignd));
     }
    }
    else
    {
     if (!strcmp(ptr, "pc"))
     {
      ptr= strtok(NULL, " \t");
      if (!ptr)  /* Just print reg */
       printf("pc= %s\n", NumOut(pc, radix, unsignd));
      else
      {
       int temp;
       if (getNum(ptr, &temp, radix) != 1)
	printf("value must be a number\n");
       else
	pc= temp;
      }  /* !ptr */
     }  /* ptr == "pc" */
     else if (sscanf(ptr, "%d", &i) != 1)
      printf("usage: reg [num [value]]  or  reg pc [value]\n");
     else
     {
      if (i < 0 || i > 31)
       printf("Register number must be between 0 and 31\n");
      else
      {
       ptr= strtok(NULL, " \t");
       if (!ptr)  /* Just print reg */
	printf("r%02d= %s\n", i, NumOut(regs[i], radix, unsignd));
       else
       {
	int temp;
	if (getNum(ptr, &temp, radix) != 1)
	 printf("value must be a number\n");
	else
	 writeReg(temp, i);
       }  /* !ptr */
      }  /* i >= 0 && i <= 31 */
     } /* sscanf() == 1 */
    } /* reg <something> */
   } /* ptr == "reg" */
   else if (!strcmp(ptr, "b") || !strcmp(ptr, "break") ||
	    !strcmp(ptr, "rb") || !strcmp(ptr, "rbreak") ||
	    !strcmp(ptr, "wb") || !strcmp(ptr, "wbreak"))
   {
    int locn;
    int type;
    if (ptr[0] == 'b') type= FETCH_BPT;
    else if (ptr[0] == 'r') type= READ_BPT;
    else type= WRITE_BPT;
    ptr= strtok(NULL, " \t");
    if (!ptr)  /* Just print */
    {
     int i;
     breakPoint *ptr;
     for (ptr= bptLists[type], i=1; ptr; ptr= ptr->next, ++i)
      printf("%d:  %s\n", i, NumOut(ptr->locn, radix, unsignd));
    }
    else  /* Do something */
    {
     if (!strcmp(ptr, "d") || !strcmp(ptr, "del") || !strcmp(ptr, "delete"))
     {
      /* Delete */
      int num;
      ptr= strtok(NULL, " \t");
      if (!ptr || sscanf(ptr, "%d", &num) != 1)
       printf("usage: break delete <number>\n");
      else
       rmBptNo(num-1, type);  /* Internally bpts are numbered from 0 */
     }
     else /* Not delete, so must be add */
     {
      int locn;
      if (getNum(ptr, &locn, radix) != 1)
       printf("usage: break <location>\n");
      else
       addBpt(locn, type);
     }
    }
   }
   else if (!strcmp(ptr, "di") || !strcmp(ptr, "dis"))
   {
    int start, fin;
    int do_dis= 1;
    ptr= strtok(NULL, " \t");
    if (!ptr)  /* No args */
    {
     start= pc-10*4;
     if (start < 0) start= 0;
     fin= pc+10*4;
    }
    else
    {
     if (getNum(ptr, &start, radix) != 1)
     {
      printf("usage: dis <start location> [<end location>]\n");
      do_dis= 0;  /* Don't dissasemble */
     }
     else  /* getNum(&start) == 1 */
     {
      fin= start + 20 * 4;
      ptr= strtok(NULL, " \t");
      if (ptr)
       if (getNum(ptr, &fin, radix) != 1)
       {
	printf("usage: dis <start location> [<end location>]\n");
	do_dis= 0;
       } /* getNum(&fin) != 1 */
     } /* getNum(&start) == 1 */
    } /* ptr */
    if (do_dis)
    {
     int i, instr;
     for (i=start; i < fin; i += 4)
     {
      if (FetchMem(&instr, i) == 1)  /* Error reading memory */
      {
       printf("Error reading location %s\n", NumOutNoPad(i, 1, 0xffffffff, radix));
       break;
      }
      printf("%s: %s\n", NumOut(i, radix, unsignd), dis(instr, radix));
     } /* for */
    } /* do_dis */
   } /* ptr == "dis" */
   else if (ptr[0] == 'd' &&
	    (ptr[1] == 'b' || ptr[1] == 'h' || ptr[1] == 'w'))
   {
    extern int terminal_width;
    int width, length;
    int start, fin;
    int do_dump= 1;
    int i;
    char type= ptr[1];
    switch (ptr[1])
    {
     case 'b': width= 1; break;
     case 'h': width= 2; break;
     case 'w': width= 4; break;
    }
    ptr= strtok(NULL, " \t");
    if (!ptr)  /* No args */
    {
     printf("usage: d%c <start location> [<end location>]\n", type);
     do_dump= 0;  /* Don't dump */
    }
    else
    {
     if (getNum(ptr, &start, radix) != 1)
     {
      printf("usage: d%c <start location> [<end location>]\n", type);
      do_dump= 0;  /* Don't dump */
     }
     else  /* getNum(&start) == 1 */
     {
      fin= start + 24 * 4;
      ptr= strtok(NULL, " \t");
      if (ptr)
       if (getNum(ptr, &fin, radix) != 1)
       {
	printf("usage: d%c <start location> [<end location>]\n", type);
	do_dump= 0;
       } /* getNum(&fin) != 1 */
     } /* getNum(&start) == 1 */
    } /* ptr */
    if (do_dump)
    {
     int length, howmany, locn, i;
     int bad_read= 0;
     /* Work out how long (in chars) each number will be */
     if (radix == 'x') length= width*2;
     else if (radix == 'd')
      length= (width == 1 ? 3 : (width == 2 ? 5 : 10)) + !unsignd;
     /* 1 extra for sign */
     else if (radix == 'o')
      length= (width == 4 ? 11 : width * 3);
     howmany= (terminal_width - (radix == 'x' ? 8 :
				 (radix == 'o' ? 11 :
				  (unsignd ? 11 : 10))))
      / (length + 1 + width);
     /*
      * Add 1 for the space after the number, <width> for the chars in the
      * ascii column at the side 
      */
     {
      /* Round down to nearest power of 2 */
      int powr= 2;
      printf("HM=%d  ", howmany);
      for (;;)
      {
       if (howmany > powr) howmany &= ~(powr-1);
       else break;
       powr <<= 1;
      }
      printf("Howmany=%d\n", howmany);
      howmany *= width;  /* Print this number of *bytes* */
     }
     for (locn= start; !bad_read && locn < fin; locn += howmany)
     {
      for (i= 0; i < howmany && locn + i < fin; i += width)
      {
       int data;
       bad_read= RMem(&data, locn + i, width);
       if (bad_read)
       {
	printf("<error reading data>\n");
	break;
       }
       if (radix == 'd') printf(unsignd ? "%0*u " : "%0*d ", length, data);
       else if (radix == 'o') printf("%0*o ", length, data);
       else printf("%0*x ", length, data);
      }
      for (i= 0; i < howmany && locn + i < fin; ++i)
      {
       int data= 0;
       bad_read= RMem(&data, locn + i, 1);  /* Just 1 byte */
       if (bad_read)
       {
	printf("<error>\n");
	break;
       }
       if (data >= ' ' && data <= '~')
	putchar(data);
       else
	putchar('.');
      }
      putchar('\n');
     }  /* for locn= start... */
    }  /* if (do_dump) */
   }  /* ptr == d<x> */
   else if (!strcmp(ptr, "ra") || !strcmp(ptr, "radix"))
   {
    ptr= strtok(NULL, " \t");
    if (!ptr || (*ptr != 'd' && *ptr != 'x' && *ptr != 'o' && *ptr != 'u'))
     printf("usage: radix d|x|o\n");
    else
    {
     if (*ptr == 'u')
     {
      radix= 'd';
      unsignd= 1;
     }
     else
     {
      radix= *ptr;
      unsignd= 0;
     }
    } /*else */
   }  /* ptr == "ra" */
   else if (!strcmp(ptr, "q") || !strcmp(ptr, "quit"))
    exit(0);
   else if (!strcmp(ptr, "h") || !strcmp(ptr, "?") || !strcmp(ptr, "help"))
   {
    puts("\nCommands:");
    puts("  c[ont]  - continue execution");
    puts("  s[tep]  - single step");
    puts("  r[eg]     - display all registers");
    puts("  r[eg] num - display register \"num\"");
    puts("  r[eg] pc  - display pc");
    puts("  r[eg] num value - set register \"num\" to \"value\"");
    puts("  r[eg] pc  value - set pc to \"value\"");
    puts("  b[reak] - list breakpoints");
    puts("  b[reak] location - set breakpoint at \"location\"");
    puts("  b[reak] delete number - delete breakpoint number \"number\"");
    puts("  rbreak - same as break, but for read points");
    puts("  wbreak - same as break, but for write points");
    puts("  di[s] - dissasemble 20 instructions around pc");
    puts("  di[s] start [end] - disassemble memory from \"start\" to \"end\"");
    puts("  db start [end] - dump memory in bytes from \"start\" to \"end\"");
    puts("  dh start [end] - dump memory in half-words from \"start\" to \"end\"");
    puts("  dw start [end] - dump memory in words from \"start\" to \"end\"");
    puts("  ra[dix] radix - sets radix: x for hex, o for octal,");
    puts("                  d for signed decimal, u for unsigned decimal");
    puts("  h[elp]");
    puts("  q[uit]");
   }
   else
    printf("Not a valid command.  Enter \"help\" for help.\n");
  } /* !using_x */
#ifndef SIM_MAIN
  else
  {
   char ch;
   int type;
   int locn;
   Read(comms_sock[1], &ch, 1);
   if (ch == 'r') break;  /* Just run */
   else if (ch == 's')
   {
    gotInt= 1;  /* Pretend we got an int. straight away */
    break;
   }
   switch (ch)
   {
    case 'b':
     Read(comms_sock[1], &ch, 1);
     if (ch == 'r') type= READ_BPT;
     else if (ch == 'w') type= WRITE_BPT;
     else if (ch == 'f') type= FETCH_BPT;
     else
     {
      message("PANIC: Add_bpt: Invalid bpt type %c", ch);
      exit(1);
     }
     Read(comms_sock[1], &locn, sizeof(locn));
     addBpt(locn, type);
     break;

    case 'd':
     Read(comms_sock[1], &ch, 1);
     if (ch == 'r') type= READ_BPT;
     else if (ch == 'w') type= WRITE_BPT;
     else if (ch == 'f') type= FETCH_BPT;
     else
     {
      message("PANIC: Del_bpt: Invalid bpt type %c", ch);
      exit(1);
     }
     Read(comms_sock[1], &locn, sizeof(locn));
     rmBpt(locn, type);
     break;

    case 'c':
     clearAllBpts();
     break;

    case 'R':
     Read(comms_sock[1], &radix, 1);
     break;

    default:
     message("PANIC: Invalid command %c", ch);
     exit(1);
   }  /* switch */
  }  /* using_x */
#endif
 } /* for */
 trapNo= -1;
 signal(ATTN_SIG, handler);  /* Restore signal handler */
 return;
}
